/*
 * Test for the new frame buffer API
 * (c) 1999 Fabrice Bellard
 */
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <linux/fb.h>
#include <termios.h>
#include <getopt.h>
#include "fbtools.h"

/* XXX: works in depth 16 only */
/* XXX: restart FBIOACCEL_EXEC system call if signal */

/* number of pixel to draw */
#define NB_PIXELS          2000000
#define XSIZE              512
#define YSIZE              256

/* maximum size of a kernel request */
#define REQUEST_MAX_SIZE   10000

/* define it to test the clipping */
#define TEST_CLIP

/* define it to measure the overhead of the display list building */
/*#define NO_SYSCALL*/

void putpixel_user(void)
{
    int i,x,y,linesize,seed;
    char *ptr;
    linesize=fb_var.xres_virtual*2;

    if (fb_var.bits_per_pixel != 16 ||
        fb_var.yres_virtual < YSIZE ||
        fb_var.xres_virtual < XSIZE) {
        fprintf(stderr,"Must be in 16 bpp to do the test\n");
        return;
    }
    seed=0;
    ptr=fb_mem;
    for(i=0;i<NB_PIXELS;i++) {
        x=i & (XSIZE-1);
        y=(i >> 9) & (YSIZE-1);
#ifdef TEST_CLIP
        if (x >= 0 && y >= 0 && x < 386 && y < 256)
#endif
            {
                *(unsigned short *)(ptr+y*linesize+x*2)=0xffff;
            }
    }
}


int fb_handle;

void putpixel_kernel(void)
{
    int x,y,i;
    fba_req_t req_buf[REQUEST_MAX_SIZE],*p;
    struct fb_accel_list disp_list;

    p=req_buf;
    FBA_SetFG(p,0xffff); /* set white color */
#ifdef TEST_CLIP
    FBA_SetClipRect(p,0,0,384,256);
#endif

    for(i=0;i<NB_PIXELS;i++) {
        x=i & (XSIZE-1);
        y=(i >> 9) & (YSIZE-1);
        FBA_Point(p,x,y);
        if ((p - req_buf) > (REQUEST_MAX_SIZE - FBA_OP_MAX_SIZE)) {
            /* request long enough: we send it */
            disp_list.list_ptr=req_buf;
            disp_list.list_size=p - req_buf;
#ifndef NO_SYSCALL
            ioctl(fb_handle,FBIOACCEL_EXEC,&disp_list);
#endif
            p=req_buf;
        }
    }
    /* last request */
#ifdef TEST_CLIP
    FBA_SetClipRect(p,0,0,fb_var.xres_virtual,fb_var.yres_virtual);
#endif
    if ((p - req_buf) > 0) {
        disp_list.list_ptr=req_buf;
        disp_list.list_size=p - req_buf;
#ifndef NO_SYSCALL
        ioctl(fb_handle,FBIOACCEL_EXEC,&disp_list);
#endif
    }
}

#define NB_GLYPHS 100000
#define GLYPH_XSIZE 8
#define GLYPH_YSIZE 13

void glyph_test(void)
{
    unsigned char glyph[GLYPH_YSIZE];
    fba_req_t req_buf[REQUEST_MAX_SIZE],*p;
    struct fb_accel_list disp_list;
    int x,y,i,tmp;

    for(i=0;i<GLYPH_YSIZE;i++) glyph[i]=0x80 >> (i & 7);

    p=req_buf;
    FBA_SetClipRect(p,10,50,600,256);

    for(i=0;i<NB_GLYPHS;i++) {
        tmp=random() & 0xffff;
        FBA_SetFG(p,tmp);

        tmp=random() & 0xffff;
        FBA_SetBG(p,tmp);

        x=random() % fb_var.xres_virtual;

        y=random() % fb_var.yres_virtual;
        
        FBA_GlyphBlt(p,x,y,GLYPH_XSIZE,GLYPH_YSIZE,glyph);
        if ((p - req_buf) > (REQUEST_MAX_SIZE - FBA_OP_MAX_SIZE*2)) {
            /* request long enough: we send it */
            disp_list.list_ptr=req_buf;
            disp_list.list_size=p - req_buf;
            ioctl(fb_handle,FBIOACCEL_EXEC,&disp_list);
            p=req_buf;
        }
    }

    FBA_SetClipRect(p,0,0,fb_var.xres_virtual,fb_var.yres_virtual);

    if ((p - req_buf) > 0) {
        disp_list.list_ptr=req_buf;
        disp_list.list_size=p - req_buf;
        ioctl(fb_handle,FBIOACCEL_EXEC,&disp_list);
        p=req_buf;
    }
}

#define NB_RECT 1000

void fillrect_test(void)
{
    fba_req_t req_buf[REQUEST_MAX_SIZE],*p;
    struct fb_accel_list disp_list;
    int x,y,xsize,ysize,i,tmp;

    p=req_buf;
    FBA_SetClipRect(p,10,50,600,256);

    for(i=0;i<NB_RECT;i++) {
        tmp=random() & 0xffff;
        FBA_SetFG(p,tmp);

        x=random() % fb_var.xres_virtual;
        y=random() % fb_var.yres_virtual;
        xsize=random() % fb_var.xres_virtual;
        ysize=random() % fb_var.yres_virtual;
        
        FBA_FillRect(p,x,y,xsize,ysize,FBA_ROP_COPY);
        if ((p - req_buf) > (REQUEST_MAX_SIZE - FBA_OP_MAX_SIZE*2)) {
            /* request long enough: we send it */
            disp_list.list_ptr=req_buf;
            disp_list.list_size=p - req_buf;
            ioctl(fb_handle,FBIOACCEL_EXEC,&disp_list);
            p=req_buf;
        }
    }

    FBA_SetClipRect(p,0,0,fb_var.xres_virtual,fb_var.yres_virtual);

    if ((p - req_buf) > 0) {
        disp_list.list_ptr=req_buf;
        disp_list.list_size=p - req_buf;
        ioctl(fb_handle,FBIOACCEL_EXEC,&disp_list);
        p=req_buf;
    }
}

char *saved_screen;
int saved_screen_size;

/* wait for a key and handle the console switch synchronously */
int wait_key(void)
{
    int len;
    unsigned char ch;

    for(;;) {
        len=read(tty,&ch,1);
        if (len == 1) break;
        switch(fb_switch_state) {
        case FB_REL_REQ:
            /* we are requested to release the display : we save the screen */
            saved_screen=malloc(saved_screen_size);
            if (saved_screen != NULL) {
                memcpy(saved_screen,fb_mem,saved_screen_size);
            }
            fb_switch_release();
            break;
        case FB_ACQ_REQ:
            /* we are request to acquire the display : copy the screen */
            if (saved_screen != NULL) {
                memcpy(fb_mem,saved_screen,saved_screen_size);
                free(saved_screen);
            }
            fb_switch_acquire();
            break;
        }
    }
    return ch;
}

long long get_time(void)
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return (long long) tv.tv_sec * 1000000LL + (long long ) tv.tv_usec;
}


void print_help(void)
{
    printf("Test program for the new frame buffer API\n"
           "(c) 1999 Fabrice Bellard\n"
           "\n"
           "usage: fbcon-test [-d fb_device] [-m fb_mode] test_type\n"
           "test_type is:\n"
           "p: putpixel in user mode\n"
           "a: putpixel in kernel mode\n"
           "g: FBA_GlyphBlt test\n"
           "r: FBA_FillRect test\n"
           );
}

int main(int argc,char **argv)
{
    char *device=NULL;
    char *mode=NULL;
    int c,bpp;
    long long start,end;

    for(;;) {
        c=getopt(argc,argv,"m:d:h");
        if (c == -1) break;
        switch(c) {
        case 'm':
            mode=optarg;
            break;
        case 'd':
            device=optarg;
            break;
        case 'h':
            print_help();
            exit(1);
        default:
            exit(1);
        }
    }

    if (optind >= argc) {
        print_help();
        exit(1);
    }

    fb_handle=fb_init(device,mode);
    if (fb_handle < 0) {
        fprintf(stderr,"Error while initializing graphic mode\n");
        exit(1);
    }
    fb_switch_init();
    atexit(tty_cleanup);
    tty_init();

    /* compute the size needed to store the screen during a console switch */
    bpp=fb_var.bits_per_pixel;
    if (bpp == 15) bpp=16;

    saved_screen_size=(fb_var.xres_virtual * fb_var.yres_virtual * bpp) >> 3;

    start=get_time();
    switch(argv[optind][0]) {
    case 'p':
        putpixel_user();
        break;
    case 'a':
        putpixel_kernel();
        break;
    case 'g':
        glyph_test();
        break;
    case 'r':
        fillrect_test();
        break;
    }
    end=get_time();

    wait_key();

    fb_cleanup();

    printf("total time=%0.6f seconds \n",(double) (end - start) / 1000000.0);

    return 0;
}
